/*--------------------------------------------------------------------------

   Copyright (c) 2006-2007, Intellectual Reserve, Inc. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 2 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  --------------------------------------------------------------------------*/

/** Implements the DNX communications methods.
 *
 * @file dnxClntProt.c
 * @author Robert W. Ingraham (dnx-devel@lists.sourceforge.net)
 * @attention Please submit patches to http://dnx.sourceforge.net
 * @ingroup DNX_CLIENT_IMPL
 */

#include "dnxClntProt.h"
#include "dnxXml.h"
#include "dnxError.h"
#include "dnxTransport.h"
#include <assert.h>
#include <string.h>

//----------------------------------------------------------------------------

/** Request a job from the registrar (client).
 *
 * The request can only be sent to the connected address on the channel. The
 * timeout is forever, which means this function will block till the entire
 * message is sent.
 *
 * @param[in] channel - the channel from which to receive the job request.
 * @param[out] pReg - the address of storage into which the request should
 *    be read from @p channel.
 *
 * @return Zero on success, or a non-zero error value.
 */
int dnxSendNodeRequest(DnxChannel * channel, DnxNodeRequest * pReg)
{
   DnxXmlBuf xbuf;

   assert(channel && pReg);

   // create the XML message
   dnxXmlOpen (&xbuf, "NodeRequest");
   dnxXmlAdd  (&xbuf, "XID",     DNX_XML_XID,  &pReg->xid);
   dnxXmlAdd  (&xbuf, "GUID",    DNX_XML_XID,  &pReg->xid);    // old format - for bc
   dnxXmlAdd  (&xbuf, "ReqType", DNX_XML_INT,  &pReg->reqType);
   dnxXmlAdd  (&xbuf, "JobCap",  DNX_XML_UINT, &pReg->jobCap);
   dnxXmlAdd  (&xbuf, "Capacity",DNX_XML_UINT, &pReg->jobCap); // old format - for bc
   dnxXmlAdd  (&xbuf, "TTL",     DNX_XML_UINT, &pReg->ttl);
   dnxXmlClose(&xbuf);

   dnxDebug(4, "dnxSendNodeRequest: XML msg(%d bytes)=%s.", xbuf.size, xbuf.buf);

   // send it on the specified channel
   return dnxPut(channel, xbuf.buf, xbuf.size, 0, 0);
}

//----------------------------------------------------------------------------

/** Wait for a job from the dispatcher (client).
 *
 * No address is returned, because only one remote peer can ever send to a 
 * connected socket.
 *
 * @param[in] channel - the channel from which to receive the job request.
 * @param[out] pJob - the address of storage into which the job request
 *    should be read from @p channel.
 * @param[in] timeout - the maximum number of seconds the caller is willing to
 *    wait before accepting a timeout error.
 *
 * @return Zero on success, or a non-zero error value.
 */
int dnxWaitForJob(DnxChannel * channel, DnxJob * pJob, int timeout)
{
   DnxXmlBuf xbuf;
   int ret;

   assert(channel && pJob);
   memset(pJob, 0, sizeof *pJob);

   // await a message from the specified channel
   xbuf.size = sizeof xbuf.buf - 1;
   if ((ret = dnxGet(channel, xbuf.buf, &xbuf.size, timeout, 0)) != DNX_OK)
      return ret;

   // decode the XML message
   xbuf.buf[xbuf.size] = 0;
   dnxDebug(4, "dnxWaitForJob: XML msg(%d bytes)=%s.", xbuf.size, xbuf.buf);

   // verify this is a "Job" message
   if ((ret = dnxXmlCmpStr(&xbuf, "Request", "Job")) != DNX_OK)
      return ret;

   // decode the job's XID (support older GUID format as well)
   if ((ret = dnxXmlGet(&xbuf, "XID", DNX_XML_XID, &pJob->xid)) != DNX_OK
         && (ret = dnxXmlGet(&xbuf, "GUID", DNX_XML_XID, &pJob->xid)) != DNX_OK)
      return ret;

   // decode the job's state
   if ((ret = dnxXmlGet(&xbuf, "State", DNX_XML_INT, &pJob->state)) != DNX_OK)
      return ret;

   // decode the job's priority
   if ((ret = dnxXmlGet(&xbuf, "Priority", DNX_XML_INT, &pJob->priority)) != DNX_OK)
      return ret;

   // decode the job's timeout
   if ((ret = dnxXmlGet(&xbuf, "Timeout", DNX_XML_INT, &pJob->timeout)) != DNX_OK)
      return ret;

   // decode the job's command
   return dnxXmlGet(&xbuf, "Command", DNX_XML_STR, &pJob->cmd);
}

//----------------------------------------------------------------------------

/** Report a job result to the collector (client).
 *
 * The result is sent to the peer the socket is connected to.
 *
 * @param[in] channel - the channel on which to send @p pResult.
 * @param[in] pResult - the result data to be sent on @p channel.
 *
 * @return Zero on success, or a non-zero error value.
 */
int dnxSendResult(DnxChannel * channel, DnxResult * pResult)
{
   DnxXmlBuf xbuf;
   char * resData;

   assert(channel && pResult);

   if ((resData = pResult->resData) == 0 || *resData == 0)
      resData = "(DNX: No Output!)";

   // create the XML message
   dnxXmlOpen (&xbuf, "Result");
   dnxXmlAdd  (&xbuf, "XID",        DNX_XML_XID,  &pResult->xid);
   dnxXmlAdd  (&xbuf, "GUID",       DNX_XML_XID,  &pResult->xid); // old format - for bc
   dnxXmlAdd  (&xbuf, "State",      DNX_XML_INT,  &pResult->state);
   dnxXmlAdd  (&xbuf, "Delta",      DNX_XML_UINT, &pResult->delta);
   dnxXmlAdd  (&xbuf, "ResultCode", DNX_XML_INT,  &pResult->resCode);
   dnxXmlAdd  (&xbuf, "ResultData", DNX_XML_STR,   resData);
   dnxXmlClose(&xbuf);

   dnxDebug(4, "dnxSendResult: XML msg(%d bytes)=%s.", xbuf.size, xbuf.buf);

   // send it on the specified channel
   return dnxPut(channel, xbuf.buf, xbuf.size, 0, 0);
}

//----------------------------------------------------------------------------

