/*--------------------------------------------------------------------------

   Copyright (c) 2006-2007, Intellectual Reserve, Inc. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 2 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  --------------------------------------------------------------------------*/

/** Implements the DNX Client logging functions.
 *
 * @file pfopen.c
 * @author Robert W. Ingraham (dnx-devel@lists.sourceforge.net)
 * @author John Calcote (dnx-devel@lists.sourceforge.net)
 * @attention Please submit patches to http://dnx.sourceforge.net
 * @ingroup DNX_CLIENT_IMPL
 */

#include "pfopen.h"

#include "dnxDebug.h"

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/wait.h>
#include <assert.h>

#define elemcount(x) (sizeof(x)/sizeof(*(x)))

//----------------------------------------------------------------------------

/** Open a child process whose STDIO is redirected to the parent.
 *
 * The STDOUT, STDERR, and (optionally) STDIN file handles for the child 
 * process are available to the parent (the calling process) to be used in 
 * any appropriate manner.
 *
 * The only allowable modes for the pipe are 'read' and 'write' as 
 * specified by the @p type parameter in the form of a single 'r' or 'w' 
 * character string. A 'r' channel is read-only and can only be used to 
 * gather data from the child process. A 'w' channel can also be used to 
 * send data to the child process.
 *
 * We use (up to) three pipes for communication. According to POSIX.1g 
 * it's not specified whether you can write to the read end of a pipe, or 
 * vice-versa; pipes are defined as half-duplex. Thus, we need to open a 
 * separate pipe (pair of file descriptors) for each channel. While full-
 * duplex pipes are probably available everywhere, it's not in the spec, so
 * it's not portable.
 *
 * <ul>
 *   <li>First Pipe:</li>
 *   <ul>
 *     <li>pfd[0] The parent reads stdout data from the child.</li>
 *     <li>pfd[1] The child writes stdout data to the parent.</li>
 *   </ul>
 *   <li>Second Pipe:</li>
 *   <ul>
 *     <li>pfd[2] The parent reads stderr data from the child.</li>
 *     <li>pfd[3] The child writes stderr data to the parent.</li>
 *   </ul>
 *   <li>Third Pipe (optional - *type == 'w'):</li>
 *   <ul>
 *     <li>pfd[4] The child reads stdin data from the parent.</li>
 *     <li>pfd[5] The parent writes stdin data to the child.</li>
 *   </ul>
 * </ul>
 *
 * @param[in] cmdstring - the command string executed as a child process.
 * @param[in] type - the file type string used when opening the I/O channel
 *    to the child process.
 *
 * @return A newly allocated PFILE object reference. This object may be
 *    queried or closed using the other calls defined in this module. If this
 *    routine returns 0, it indicates an error and errno is set appropriately.
 */
PFILE * pfopen(const char * cmdstring, const char * type)
{
   PFILE * pfile;
   int pid, eno, i, pfd[6];

   // only allow "r" or "w"
   assert((type[0] == 'r' || type[0] == 'w') &&  type[1] == 0);

   // Allocate a PFILE structure
   if ((pfile = (PFILE *)xcalloc(1, sizeof *pfile)) == 0)
   {
      errno = ENOMEM;
      return 0;
   }

   // set all descriptors to unallocated state and open pipes
   memset(pfd, -1, sizeof pfd);
   if (pipe(pfd) < 0 || pipe(pfd+2) < 0
         || (*type == 'w' && pipe(pfd+4) < 0))
      goto errout;

   // create buffered channels for parent ends of pipes
   if ((pfile->fp[0] = fdopen(pfd[0], "r")) == 0
         || (pfile->fp[1] = fdopen(pfd[2], "r")) == 0
         || (*type == 'w' && (pfile->fp[2] = fdopen(pfd[5], "w")) == 0))
      goto errout;

   // fork and check for error
   if ((pid = fork()) < 0) 
      goto errout;

   // child process?
   if (pid == 0)
   {
      // make child its own process group so parent can signal child subtree
      setpgid(0, 0);

      // close parent (read) ends, dup child (write) ends (if needed)
      fclose(pfile->fp[0]);
      fclose(pfile->fp[1]);
      if (pfd[1] != STDOUT_FILENO) 
      {
         dup2(pfd[1], STDOUT_FILENO);
         close(pfd[1]);
      }
      if (pfd[3] != STDERR_FILENO) 
      {
         dup2(pfd[3], STDERR_FILENO);
         close(pfd[3]);
      }

      // if also sending input from parent to child...
      if (*type == 'w')
      {
         // close parent (write) end, dup child (read) end (if needed)
         fclose(pfile->fp[2]);
         if (pfd[4] != STDIN_FILENO) 
         {
            dup2(pfd[4], STDIN_FILENO);
            close(pfd[4]);
         }
      } 

      execl("/bin/sh", "/bin/sh", "-c", cmdstring, (char *)0);

      _exit(127); // if we get here, we got an exec error
   }

   // parent continues in this process...

   // close child ends of pipes
   close(pfd[1]);
   close(pfd[3]);
   if (*type == 'w')
      close(pfd[4]);

   // save child process pid
   pfile->pid = pid;

   return pfile;

errout:
   eno = errno;  // save errno (close changes it)

   // close buffered parent FILEs (also closes descriptors)
   if (pfile->fp[0] != 0)
      fclose(pfile->fp[0]), pfd[0] = -1;
   if (pfile->fp[1] != 0)
      fclose(pfile->fp[1]), pfd[2] = -1;
   if (pfile->fp[2] != 0)
      fclose(pfile->fp[2]), pfd[5] = -1;

   // close any remaining unbuffered descriptors
   for (i = 0; i < elemcount(pfd); i++)
      if (pfd[i] >= 0)
         close(pfd[i]);

   xfree(pfile);
   errno = eno;
   return 0;
}

//----------------------------------------------------------------------------

/** Close an existing PFILE object.
 *
 * Closes all open pipe handles, and then waits for the associated child
 * process to die, returning the status value of the child process.
 *
 * @param[in] pfile - the process pipe to be closed.
 *
 * @return The value of the waitpid system call's stat_loc parameter, or the
 *    status of the child processed waited for using the waitpid system call.
 *
 * @note Because the return value of this routine is the stat_loc value of
 *    the waitpid system call, any of the W* macros defined for use on this
 *    value by the system may be used to query the status of the associated
 *    child process.
 */
int pfclose(PFILE * pfile)
{
   int stat;

   if (pfile == 0)
   {
      errno = EINVAL;
      return -1;     /* popen() has never been called */
   }

   if (pfile->fp[0] != 0)
      fclose(pfile->fp[0]);
   if (pfile->fp[1] != 0)
      fclose(pfile->fp[1]);
   if (pfile->fp[2] != 0)
      fclose(pfile->fp[2]);
   
   xfree(pfile);

   while (waitpid(pfile->pid, &stat, 0) < 0)
      if (errno != EINTR)
         return -1;  /* error other than EINTR from waitpid() */
   
   return stat;      /* return child's termination status */
}

//----------------------------------------------------------------------------

/** Kill a set of processes associated with an existing PFILE object.
 *
 * @param[in] pfile - the pipe representing the process to be killed.
 * @param[in] sig - the signal to send the process.
 *
 * @return Zero on success, or -1 on error, and the global errno is set.
 */
int pfkill(PFILE * pfile, int sig)
{
   assert(pfile);

   /* send specified signal to child process group */
   return kill(-(pfile->pid), sig);
}

/*--------------------------------------------------------------------------
                                 TEST MAIN

   From within dnx/client, compile with GNU tools using this command line:

      gcc -DDEBUG -DPFOPEN_TEST -g -O0 -o pfopenTest -I../common \
         ../common/dnxError.c pfopen.c

   Alternatively, a heap check may be done with the following command line:

      gcc -DDEBUG -DDEBUG_HEAP -DPFOPEN_TEST -g -O0 -o pfopenTest \
         -I../common ../common/dnxError.c pfopen.c

  --------------------------------------------------------------------------*/

#ifdef PFOPEN_TEST

#include "utesthelp.h"
#include "dnxError.h"

// Implementation Note: m4 echos back on STDOUT whatever it reads on
// STDIN; when killed (SIGTERM), it writes "Terminated\n" to STDERR.

char * program = "m4";
char * text = "This is a test\n";
char * errtext = "Terminated\n";
int verbose;

int main(int argc, char ** argv)
{
   PFILE * pf;
   FILE * fin, * fout, * ferr;
   char buf[128];

   verbose = argc > 1 ? 1 : 0;

   CHECK_NONZERO(pf = pfopen(program, "w"));

   fin  = PF_IN(pf);
   fout = PF_OUT(pf);
   ferr = PF_ERR(pf);

   fprintf(fin, text);
   fflush(fin);
   close(fileno(fin));

   fgets(buf, sizeof buf, fout);

   CHECK_ZERO(strcmp(text, buf));

   CHECK_ZERO(pfclose(pf));

#ifdef DEBUG_HEAP
   CHECK_ZERO(dnxCheckHeap());
#endif

   return 0;
}

#endif   /* PFOPEN_TEST */

/*--------------------------------------------------------------------------*/

