/*--------------------------------------------------------------------------

   Copyright (c) 2006-2007, Intellectual Reserve, Inc. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 2 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  --------------------------------------------------------------------------*/

/** Implements the DNX UDP Transport Layer.
 *
 * @file dnxUdp.c
 * @author Robert W. Ingraham (dnx-devel@lists.sourceforge.net)
 * @attention Please submit patches to http://dnx.sourceforge.net
 * @ingroup DNX_COMMON_IMPL
 */

#include "dnxUdp.h"
#include "dnxTSPI.h"

#include "dnxTransport.h"
#include "dnxError.h"
#include "dnxDebug.h"
#include "dnxLogging.h"

#include "dnxProtocol.h"

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <pthread.h>
#include <limits.h>

#ifndef HOST_NAME_MAX
# define HOST_NAME_MAX 256
#endif

/** The implementation of the UDP low-level I/O transport. */
typedef struct iDnxUdpChannel_
{
   char * host;         //!< Channel transport host name.
   int port;            //!< Channel transport port number.
   int socket;          //!< Channel transport socket.
   DnxTransStats stats; //!< Channel transport statistics.
   iDnxChannel ichan;   //!< Channel transport I/O (TSPI) methods.
} iDnxUdpChannel;

/** @todo Use GNU reentrant resolver extensions on platforms where available. */

static pthread_mutex_t udpMutex;

/*--------------------------------------------------------------------------
                  TRANSPORT SERVICE PROVIDER INTERFACE
  --------------------------------------------------------------------------*/

/** Open a UDP channel object.
 *
 * If DNX_MODE_ACTIVE is passed in @p mode, the channel is considered a
 * client-side connection, and the address stored in @p icp is interpreted
 * as a remote peer address to be connected to. Active mode connections
 * must provide a resolvable internet address.
 *
 * If DNX_MODE_PASSIVE is passed in @p mode, the channel is considered a 
 * server-side binding, and the address stored in @p icp is interpreted as
 * a local address to be bound to. Passive mode connections may use the
 * address of a specific interface on the local host, or the values "0",
 * "0.0.0.0", or "INADDR_ANY" to specify that all interfaces should be 
 * bound on this connection (the actual local address of a received message
 * is determined by which interface actually receives it.
 *
 * @param[in] icp - the UDP channel object to be opened.
 * @param[in] mode - DNX_MODE_ACTIVE indicates the transport will be 
 *    used in active mode (as a client); DNX_MODE_PASSIVE indicates the 
 *    transport will be used in passive mode (as a server listen point).
 *
 * @return Zero on success, or a non-zero error value.
 */
static int dnxUdpOpen(iDnxChannel * icp, int mode)
{
   iDnxUdpChannel * iucp = (iDnxUdpChannel *)
         ((char *)icp - offsetof(iDnxUdpChannel, ichan));
   struct hostent * he;
   struct sockaddr_in inaddr;
   int sd;

   assert(icp && iucp->port > 0);

   inaddr.sin_family = AF_INET;
   inaddr.sin_port = htons((in_port_t)iucp->port);

   // see if we are listening on any address
   if (!strcmp(iucp->host, "INADDR_ANY")
         || !strcmp(iucp->host, "0.0.0.0")
         || !strcmp(iucp->host, "0"))
   {
      // make sure that the requested mode is passive
      if (mode == DNX_MODE_ACTIVE) return DNX_ERR_ADDRESS;
      inaddr.sin_addr.s_addr = INADDR_ANY;
   }
   else
   {
      DNX_PT_MUTEX_LOCK(&udpMutex);
      if ((he = gethostbyname(iucp->host)) != 0)
         memcpy(&inaddr.sin_addr.s_addr, he->h_addr_list[0], he->h_length);
      DNX_PT_MUTEX_UNLOCK(&udpMutex);

      if (!he) return DNX_ERR_ADDRESS;
   }

   if ((sd = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
   {
      dnxLog("dnxUdpOpen: socket failed: %s.", strerror(errno));
      return DNX_ERR_OPEN;
   }

   // determine how to handle socket connectivity based upon open mode.
   if (mode == DNX_MODE_ACTIVE)
   {
      struct sockaddr_in inaddr_any;

      // want to sometimes listen for incoming packets on this 
      // active channel also, so bind socket to any local address
      // (if not done, would happen anyway on first packet sent)
      memset(&inaddr_any, 0, sizeof inaddr_any);
      inaddr_any.sin_family = AF_INET;
      inaddr_any.sin_addr.s_addr = htonl(INADDR_ANY);
      if (bind(sd, (struct sockaddr *)&inaddr_any, sizeof inaddr_any) != 0)
      {
         close(sd);
         dnxLog("dnxUdpOpen: bind(any) failed: %s.", strerror(errno));
         return DNX_ERR_OPEN;
      }

      // for UDP, this sets the default destination address, so we can
      // now use send() and write() in addition to sendto() and sendmsg()
      // (also makes it impossible to specify sendto address or retrieve
      // sender address in recvfrom, but there's only one in either case)
      if (connect(sd, (struct sockaddr *)&inaddr, sizeof inaddr) != 0)
      {
         dnxLog("dnxUdpOpen: connect(%lx) failed: %s.",
               (unsigned long)inaddr.sin_addr.s_addr, strerror(errno));
         close(sd);
         return DNX_ERR_OPEN;
      }
   }
   else
   {
      struct linger lopt;
      int reuse = 1;

      lopt.l_onoff = 0;
      lopt.l_linger = 0;

      // reuse addr and clear linger to avoid tcp time_wait state issues
      setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, &reuse, sizeof reuse);
      setsockopt(sd, SOL_SOCKET, SO_LINGER, &lopt, sizeof lopt);

      // want to listen for incoming packets, so bind to the
      // specified local address and port
      if (bind(sd, (struct sockaddr *)&inaddr, sizeof inaddr) != 0)
      {
         close(sd);
         dnxLog("dnxUdpOpen: bind(%lx) failed: %s.",
               (unsigned long)inaddr.sin_addr.s_addr, strerror(errno));
         return DNX_ERR_OPEN;
      }
   }

   iucp->socket = sd;

   return DNX_OK;
}

//----------------------------------------------------------------------------

/** Close a UDP channel object.
 *
 * @param[in] icp - the UDP channel object to be closed.
 *
 * @return Always returns zero.
 */
static int dnxUdpClose(iDnxChannel * icp)
{
   iDnxUdpChannel * iucp = (iDnxUdpChannel *)
         ((char *)icp - offsetof(iDnxUdpChannel, ichan));

   assert(icp && iucp->socket);

   close(iucp->socket);
   iucp->socket = 0;

   return DNX_OK;
}

//----------------------------------------------------------------------------

/** Read data from a UDP channel object.
 *
 * Posix.1g specifies that a connected socket cannot receive the address
 * of the remote peer, and that connected sockets should only use the recv
 * call. Our architecture only gives us one way of determining if this
 * socket is connected or not: If the @p src parameter is null, this
 * socket is presumed to be an active, connected, or client-side socket 
 * being used to receive datagrams from a server in response to a prior
 * request. If @p src is NOT null, this socket is assumed to be a server-
 * side socket bound to a specific address, but unconnected. Thus it's 
 * important that higher layers pass appropriate values for @p src or
 * the wrong functions will be called.
 *
   // mlen can be zero - sending a zero-length packet is allowed, but unlikely
   // mlen can't be greater than DNX_MAX_MSG however because we only allows *size
   // bytes to be received, and the upper layers don't ask for larger than that.

 * @param[in] icp - the UDP channel object from which to read data.
 * @param[out] buf - the address of storage into which data should be read.
 * @param[in,out] size - on entry, the maximum number of bytes that may be
 *    read into @p buf; on exit, returns the number of bytes stored in @p buf.
 * @param[in] timeout - the maximum number of seconds we're willing to wait
 *    for data to become available on @p icp without returning a timeout
 *    error.
 * @param[out] src - the address of storage for the sender's address if
 *    desired. This parameter is optional, and may be passed as NULL. If
 *    non-NULL, the buffer pointed to by @p src must be at least the size
 *    of a @em sockaddr_in structure.
 *
 * @return Zero on success, or a non-zero error value.
 */
static int dnxUdpRead(iDnxChannel * icp, char * buf, int * size, int timeout, char * src)
{
   iDnxUdpChannel * iucp = (iDnxUdpChannel *)((char *)icp - offsetof(iDnxUdpChannel, ichan));
   socklen_t slen = sizeof(struct sockaddr_storage);
   ssize_t mlen;

   assert(icp && iucp->socket && buf && size && *size > 0);

   // implement timeout logic, if timeout value is greater than zero
   if (timeout > 0)
   {
      struct timeval tv;
      fd_set fd_rds;
      int nsd;

      FD_ZERO(&fd_rds);
      FD_SET(iucp->socket, &fd_rds);

      tv.tv_usec = 0L;
      tv.tv_sec = timeout;

      if ((nsd = select(iucp->socket + 1, &fd_rds, 0, 0, &tv)) == 0)
         return DNX_ERR_TIMEOUT;

      if (nsd < 0)
      {
         if (errno != EINTR)
         {
            dnxLog("UDP receive - select failed: %s.", strerror(errno));
            return DNX_ERR_RECEIVE;
         }
         return DNX_ERR_TIMEOUT;
      }
   }

   // read the incoming message, use recvfrom only if address was requested
   if (src)
      mlen = recvfrom(iucp->socket, buf, *size, 0, (struct sockaddr *)src, &slen);
   else 
      mlen = read(iucp->socket, buf, *size);

   if (mlen < 0)
   {
      iucp->stats.rderrs++;

      // on "connected" UDP sockets, if the server can't be reached -
      // ICMP returned ICMP_UNREACH - we could get notified here.
      if (errno == ECONNREFUSED) return DNX_ERR_ADDRESS;

      dnxDebug(2, "UDP receive - %s socket failed: %s.", 
            src? "recvfrom unconnected": "read connected", strerror(errno));

      return DNX_ERR_RECEIVE;
   }

   iucp->stats.reads++;
   *size = (int)mlen;

   return DNX_OK;
}

//----------------------------------------------------------------------------

/** Write data to a UDP channel object.
 *
 * Posix.1g specifies that a connected socket cannot send to a specific
 * address. The sendto method may be used on a connected socket, but the
 * toaddr argument must be null and the toaddrsz argument is ignored. The 
 * effect of calling this method with a non-null @p dst argument on a 
 * connected socket is not clear on some platforms. Our architecture only
 * allows us to determine if a socket was created as active or passive based
 * on the value of the @p dst parameter. Thus, this function relies on higher 
 * layers properly calling this function with a null @p dst argument for 
 * active, connected, client-side sockets, and with a non-null @p dst argument
 * for passive, bound, server-side sockets.
 *
 * Since this module doesn't create non-blocking sockets, writes and reads
 * will block until timeout occurs, or until all bytes are read or written.
 * Thus, it's not possible to send more than zero, but less than requested.
 *
 * @param[in] icp - the UDP channel object on which to write data.
 * @param[in] buf - a pointer to the data to be written.
 * @param[in] size - the number of bytes to be written on @p icp.
 * @param[in] timeout - the maximum number of seconds to wait for the write
 *    operation to complete without returning a timeout error.
 * @param[in] dst - the address to which the data in @p buf should be sent
 *    using this channel. This parameter is ignored for virtual connection
 *    based channels. This parameter is optional and may be passed as NULL.
 *
 * @return Zero on success, or a non-zero error value.
 *
 * @note If this is a stream oriented channel, or if NULL is passed for
 * the @p dst parameter, The channel destination address is used.
 */
static int dnxUdpWrite(iDnxChannel * icp, char * buf, int size, int timeout, char * dst)
{
   iDnxUdpChannel * iucp = (iDnxUdpChannel *) ((char *)icp - offsetof(iDnxUdpChannel, ichan));
   ssize_t mlen;

   assert(icp && iucp->socket && buf && size);

   // implement timeout logic, if timeout value is greater than zero
   if (timeout > 0)
   {
      struct timeval tv;
      fd_set fd_wrs;
      int nsd;

      FD_ZERO(&fd_wrs);
      FD_SET(iucp->socket, &fd_wrs);

      tv.tv_usec = 0L;
      tv.tv_sec = timeout;

      if ((nsd = select(iucp->socket + 1, 0, &fd_wrs, 0, &tv)) == 0)
         return DNX_ERR_TIMEOUT;
      if (nsd < 0)
      {
         if (errno != EINTR)
         {
            dnxLog("UDP send - select failed: %s.", strerror(errno));
            return DNX_ERR_SEND;
         }
         return DNX_ERR_TIMEOUT;
      }
   }

   // specify destination address for passive, server-side sockets only
   if (dst)
   {
      struct sockaddr * sa = (struct sockaddr *)dst;
      mlen = sendto(iucp->socket, buf, size, 0, sa, sa->sa_family == AF_INET?
            sizeof(struct sockaddr_in): sizeof(struct sockaddr_in6));
   }
   else
      mlen = write(iucp->socket, buf, size);

   if (mlen < 0)
   {
      iucp->stats.wrerrs++;
      dnxDebug(2, "UDP send - %s socket failed: %s.", 
            dst? "sendto unconnected": "write connected", strerror(errno));
      return DNX_ERR_SEND;
   }

   iucp->stats.writes++;

   return DNX_OK;
}

//----------------------------------------------------------------------------

/** Delete a UDP channel object.
 *
 * @param[in] icp - the UDP channel object to be deleted.
 */
static void dnxUdpDelete(iDnxChannel * icp)
{
   iDnxUdpChannel * iucp = (iDnxUdpChannel *)
         ((char *)icp - offsetof(iDnxUdpChannel, ichan));

   assert(icp && iucp->socket == 0);

   xfree(iucp->host);
   xfree(iucp);
}

//----------------------------------------------------------------------------

/** Return the read/write/error stats of this channel object.
 *
 * @param[in] icp - the channel for which stats should be returned.
 * @param[out] tsp - the stats structure to be filled.
 */
static void dnxUdpGetStats(iDnxChannel * icp, DnxTransStats * tsp)
{
   iDnxUdpChannel * iucp = (iDnxUdpChannel *)
         ((char *)icp - offsetof(iDnxUdpChannel, ichan));

   assert(icp && tsp);

   *tsp = iucp->stats;
}

//----------------------------------------------------------------------------

/** Reset the read/write/error stats on this channel object.
 *
 * @param[in] icp - the channel for which stats should be reset.
 */
static void dnxUdpResetStats(iDnxChannel * icp)
{
   iDnxUdpChannel * iucp = (iDnxUdpChannel *)
         ((char *)icp - offsetof(iDnxUdpChannel, ichan));

   assert(icp);

   memset(&iucp->stats, 0, sizeof iucp->stats);
}

//----------------------------------------------------------------------------

/** Create a new UDP transport.
 *
 * @param[in] url - the URL containing the host name and port number.
 * @param[out] icpp - the address of storage for returning the new low-
 *    level UDP transport object (as a generic transport object).
 *
 * @return Zero on success, or a non-zero error value.
 */
static int dnxUdpNew(char * url, iDnxChannel ** icpp)
{
   char * cp, * ep, * lastchar;
   iDnxUdpChannel * iucp = NULL;
   long port;

   assert(icpp && url && *url);

   // search for host:port in URL
   if ((cp = strstr(url, "://")) == 0)
      return DNX_ERR_BADURL;
   cp += 3;

   // find host/port separator ':' - copy host name
   if ((ep = strchr(cp, ':')) == 0 || ep == cp || ep - cp > HOST_NAME_MAX)
      return DNX_ERR_BADURL;

   // extract port number
   if ((port = strtol(ep + 1, &lastchar, 0)) < 1 || port > 65535
         || (*lastchar && *lastchar != '/'))
      return DNX_ERR_BADURL;

    if(iucp)
        xfree(iucp);

   // allocate a new iDnxUdpChannel object
    if ((iucp = (iDnxUdpChannel *)xcalloc(1,sizeof *iucp)) == 0)
    {
        return DNX_ERR_MEMORY;
    }


   // save host name and port
   if ((iucp->host = (char *)xmalloc(ep - cp + 1)) == 0)
   {
      xfree(iucp);
      return DNX_ERR_MEMORY;
   }
   memcpy(iucp->host, cp, ep - cp);
   iucp->host[ep - cp] = 0;
   iucp->port = (int)port;

   // set I/O methods
   iucp->ichan.txOpen         = dnxUdpOpen;
   iucp->ichan.txClose        = dnxUdpClose;
   iucp->ichan.txRead         = dnxUdpRead;
   iucp->ichan.txWrite        = dnxUdpWrite;
   iucp->ichan.txDelete       = dnxUdpDelete;
   iucp->ichan.txGetStats     = dnxUdpGetStats;
   iucp->ichan.txResetStats   = dnxUdpResetStats;
   *icpp = &iucp->ichan;

   return DNX_OK;
}

/*--------------------------------------------------------------------------
                           EXPORTED INTERFACE
  --------------------------------------------------------------------------*/

/** Initialize the UDP transport sub-system; return UDP channel contructor.
 *
 * @param[out] ptxAlloc - the address of storage in which to return the
 *    address of the UDP channel object constructor (dnxUdpNew).
 *
 * @return Always returns zero.
 */
int dnxUdpInit(int (**ptxAlloc)(char * url, iDnxChannel ** icpp))
{
   DNX_PT_MUTEX_INIT(&udpMutex);

   *ptxAlloc = dnxUdpNew;

   return DNX_OK;
}

//----------------------------------------------------------------------------

/** Clean up global resources allocated by the UDP transport sub-system.
 */
void dnxUdpDeInit(void)
{
   DNX_PT_MUTEX_DESTROY(&udpMutex);
}

/*--------------------------------------------------------------------------*/

