/*--------------------------------------------------------------------------

   Copyright (c) 2006-2010, Intellectual Reserve, Inc. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 2 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  --------------------------------------------------------------------------*/

/** Implements the split-server version of the DNX Nagios plugin.
 *
 * This version of the DNX Nagios Event Broker (NEB) module plugin library
 * only implements the most basic of services required as a Nagios plugin.
 * It starts the DNX Server as a child process, then creates a thread to 
 * listen for response data from the server. Finally, it registers a service 
 * check handler which begins to receive event service notifications.
 *
 * It intercepts all service checks and dispatches them to distributed
 * worker nodes via the DNX server child process. Communication with the 
 * child process is managed through inherited pipe file handles dup'd to 
 * well-known file descriptor numbers.
 *
 * @file dnx/plugin/dnxNebMain.c
 * @author John Calcote (dnx-devel@lists.sourceforge.net)
 * @attention Please submit patches to http://dnx.sourceforge.net
 * @ingroup DNX_PLUGIN_IMPL
 */

/*!@defgroup DNX_PLUGIN_IMPL DNX Nagios Plugin Implementation
 * @defgroup DNX_PLUGIN_IFC  DNX Nagios Plugin Interface
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include "dnxNebMain.h"

#include "dnxCfgParser.h"
#include "dnxError.h"
#include "dnxDebug.h"

#ifndef NSCORE
# define NSCORE
#endif
#include "nagios.h"
#include "objects.h"    // for nagios service data type
#include "nebmodules.h"
#include "nebstructs.h"
#include "nebcallbacks.h"
#include "neberrors.h"
#include "broker.h"

#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <regex.h>

#ifndef VERSION
# define VERSION    "<unknown>"
#endif

#ifndef SYSCONFDIR
# define SYSCONFDIR "/etc"
#endif

#ifndef SYSLOGDIR
# define SYSLOGDIR  "/var/log"
#endif

#ifndef LIBEXECDIR
# define LIBEXECDIR "/usr/libexec"
#endif

#if CURRENT_NEB_API_VERSION == 2
# define OBJECT_FIELD_NAME object
#elif CURRENT_NEB_API_VERSION == 3
# define OBJECT_FIELD_NAME object_ptr
#else
# error Unsupported NEB API version.
#endif

#define elemcount(x) (sizeof(x)/sizeof(*(x)))

#define DNX_DEFAULT_SERVER_CONFIG_FILE SYSCONFDIR "/dnxServer.cfg"
#define DNX_DEFAULT_LOG                SYSLOGDIR  "/dnxsrv.log"
#define DNX_DEFAULT_DBGLOG             SYSLOGDIR  "/dnxsrv.dbg.log"
#define DNX_DEFAULT_SERVER             LIBEXECDIR "/dnxServer"

/** specify event broker API version (required by Nagios). */
NEB_API_VERSION(CURRENT_NEB_API_VERSION);

/** The internal server module configuration data structure. */
typedef struct DnxCfgData
{
   char * serverPath;         //!< The path to the dnx server program.
   char * localCheckPattern;  //!< The regular expression for local jobs.
   char * syncScript;         //!< The sync script path and file name.
   char * logFilePath;        //!< The system log file path.
   char * debugFilePath;      //!< The debug log file path.
   char * auditFilePath;      //!< The audit log file path.
   unsigned debugLevel;       //!< The global debug level.
} DnxCfgData;

// module static data
static char * cfgfile;        //!< The server configuration file.
static DnxCfgData cfg;        //!< The server configuration parameters.
static DnxCfgParser * parser; //!< The system configuration parser.
static time_t start_time;     //!< The module start time.
static void * myHandle;       //!< Private NEB module handle.
static regex_t regEx;         //!< Compiled regular expression structure.
static pid_t serverpid;       //!< Server child process id.
static int outdatafd;         //!< Pipe write channel to server.
static int inackfd;           //!< Pipe ack/nak channel from server.
static int indatafd;          //!< Pipe read channel from server.
static pthread_t rlthread;    //!< Thread handle for results listener.

/*--------------------------------------------------------------------------
                              IMPLEMENTATION
  --------------------------------------------------------------------------*/

/** Cleanup the config file parser. */
static void releaseConfig(void)
{
   if (cfg.localCheckPattern)
      regfree(&regEx);

   dnxCfgParserDestroy(parser);
}

//----------------------------------------------------------------------------

/** Validate a configuration data structure in context.
 *
 * @param[in] dict - the dictionary used by the DnxCfgParser.
 * @param[in] vptrs - an array of opaque objects (either pointers or values)
 *    to be checked.
 * @param[in] passthru - an opaque pointer passed through from
 *    dnxCfgParserCreate. In this routine, it's the regex_t object into which
 *    we should parse the regular expression if one is given.
 *
 * @return Zero on success, or a non-zero error value. This error value is
 * passed back through dnxCfgParserParse.
 */
static int validateCfg(DnxCfgDict * dict, void ** vptrs, void * passthru)
{
   regex_t * rep = (regex_t *)passthru;
   int err, ret = DNX_ERR_INVALID;
   DnxCfgData cfg;
   struct stat statbuf;

   assert(dict && vptrs && passthru);

   // setup data structure so we can use the same functionality we had before
   cfg.serverPath         = (char *)vptrs[ 0];
   cfg.localCheckPattern  = (char *)vptrs[ 1];
   cfg.syncScript         = (char *)vptrs[ 2];
   cfg.logFilePath        = (char *)vptrs[ 3];
   cfg.debugFilePath      = (char *)vptrs[ 4];
   cfg.debugLevel         = (unsigned)(intptr_t)vptrs[ 5];
   cfg.auditFilePath      = (char *)vptrs[ 6];

   // validate configuration items in context
   if (!cfg.serverPath || stat(cfg.serverPath, &statbuf) < 0
         || !S_ISREG(statbuf.st_mode))
      dnxLog("config: server not specified or invalid (\"%s\").",
             cfg.serverPath);
   else if (cfg.localCheckPattern && (err = regcomp(rep,
         cfg.localCheckPattern, REG_EXTENDED | REG_NOSUB)) != 0)
   {
      char buffer[128];
      regerror(err, rep, buffer, sizeof buffer);
      dnxLog("config: Failed to compile localCheckPattern (\"%s\"): %s.",
             cfg.localCheckPattern, buffer);
      regfree(rep);
   }
   else
      ret = 0;

   return ret;
}

//----------------------------------------------------------------------------

/** Read and parse the dnxServer configuration file.
 *
 * @param[in] cfgfile - the configuration file to be read.
 *
 * @return Zero on success, or a non-zero error value.
 */
static int initConfig(char * cfgfile)
{
   DnxCfgDict dict[] =
   {  // Do NOT change the order, unless you know what you're doing!
      { "serverPath",         DNX_CFG_FSPATH,   &cfg.serverPath         },
      { "localCheckPattern",  DNX_CFG_STRING,   &cfg.localCheckPattern  },
      { "syncScript",         DNX_CFG_FSPATH,   &cfg.syncScript         },
      { "pluginLogFile",      DNX_CFG_FSPATH,   &cfg.logFilePath        },
      { "pluginDebugFile",    DNX_CFG_FSPATH,   &cfg.debugFilePath      },
      { "pluginDebugLevel",   DNX_CFG_UNSIGNED, &cfg.debugLevel         },
      { "pluginAuditFile",    DNX_CFG_FSPATH,   &cfg.auditFilePath      },
      { 0 },
   };
   char cfgdefs[] =
      "serverPath = " DNX_DEFAULT_SERVER "\n"
      "pluginLogFile = " DNX_DEFAULT_LOG "\n"
      "pluginDebugFile = " DNX_DEFAULT_DBGLOG "\n";

   int ret;
   regex_t re;

   // clear the regex string, as we may write into it
   memset(&re, 0, sizeof re);

   // create global configuration parser object
   if ((ret = dnxCfgParserCreate(cfgdefs, cfgfile, 0, dict,
         validateCfg, &parser)) != 0)
      return ret;

   // parse configuration file; pass defaults
   if ((ret = dnxCfgParserParse(parser, &re)) == 0)
      regEx = re;
   else
      dnxCfgParserDestroy(parser);

   return ret;
}

//----------------------------------------------------------------------------

/** Return the number of services configured in Nagios.
 *
 * This technique works fine as long as it's called after Nagios has had a 
 * chance to build the list. Since Nagios loads NEB modules before configuring
 * itself, we can't call this routine until Nagios calls our ehProcessData
 * handler.
 *
 * @return The number of services configured in Nagios.
 *
 * @todo This routine should be in nagios code.
 */
static int nagiosGetServiceCount(void)
{
   extern service * service_list;      // the global nagios service list

   service * temp_service;
   int total_services = 0;

   // walk the service list, count the nodes
   for (temp_service = service_list; temp_service;
         temp_service = temp_service->next)
      total_services++;

   return total_services;
}

#if CURRENT_NEB_API_VERSION == 2

//----------------------------------------------------------------------------

/** Post job result information to Nagios 2.x.
 *
 * @param[in] svc - the nagios service to which the results belong.
 * @param[in] start_time - the check start time in seconds.
 * @param[in] early_timeout - boolean; true (!0) means the service timed out.
 * @param[in] res_code - the check result code.
 * @param[in] res_data - the check result data.
 *
 * @return Zero on success, or a non-zero error value.
 *
 * @todo This routine should be in nagios code.
 */
static int nagios2xPostResult(service * svc, time_t start_time,
      int early_timeout, int res_code, char * res_data)
{
   extern circular_buffer service_result_buffer;
   int check_result_buffer_slots = 4096;

   service_message * new_message;

   // note that we're using malloc, not xmalloc - nagios takes ownership
   if ((new_message = (service_message *)malloc(sizeof *new_message)) == 0)
      return DNX_ERR_MEMORY;

   gettimeofday(&new_message->finish_time, 0);
   strncpy(new_message->host_name, svc->host_name, 
         sizeof(new_message->host_name) - 1);
   new_message->host_name[sizeof(new_message->host_name) - 1] = 0;
   strncpy(new_message->description, svc->description,
         sizeof(new_message->description) - 1);
   new_message->description[sizeof(new_message->description) - 1] = 0;
   new_message->return_code = res_code;
   new_message->exited_ok = TRUE;
   new_message->check_type = SERVICE_CHECK_ACTIVE;
   new_message->parallelized = svc->parallelize;
   new_message->start_time.tv_sec = start_time;
   new_message->start_time.tv_usec = 0L;
   new_message->early_timeout = early_timeout;
   strncpy(new_message->output, res_data, sizeof(new_message->output) - 1);
   new_message->output[sizeof(new_message->output) - 1] = 0;

   pthread_mutex_lock(&service_result_buffer.buffer_lock);

   // handle overflow conditions
   if (service_result_buffer.items == check_result_buffer_slots)
   {
      service_result_buffer.overflow++;
      service_result_buffer.tail = (service_result_buffer.tail + 1)
            % check_result_buffer_slots;
   }

   // save the data to the buffer
   ((service_message **)service_result_buffer.buffer)
         [service_result_buffer.head] = new_message;

   // increment the head counter and items
   service_result_buffer.head = (service_result_buffer.head + 1)
         % check_result_buffer_slots;
   if (service_result_buffer.items < check_result_buffer_slots)
      service_result_buffer.items++;
   if (service_result_buffer.items > service_result_buffer.high)
      service_result_buffer.high = service_result_buffer.items;

   pthread_mutex_unlock(&service_result_buffer.buffer_lock);

   return 0;
}

#elif CURRENT_NEB_API_VERSION == 3

// comment out this line to post to N3 filesystem results queue
#define DIRECT_POST

#ifdef DIRECT_POST

// a nagios core global (It shouldn't be global, but we're glad it is!)
extern check_result * check_result_list;

static check_result * dnxResultList = 0;
static pthread_mutex_t dnxResultListMutex = PTHREAD_MUTEX_INITIALIZER;

//----------------------------------------------------------------------------

/** Compare two timestamps.
 *
 * Compares two timestamps for relative value.
 *
 * @param[in] tv1 - the first timestamp.
 * @param[in] tv2 - the second timestamp.
 * @return a long value less than zero if tv1 is less than tv2, 0 if tv1 
 * is equal to tv2, or a long value greater than zero if tv1 is greater 
 * than tv2.
 */
static long dnxTimeCompare(struct timeval * tv1, struct timeval * tv2)
{
   long secdiff = (long)(tv1->tv_sec - tv2->tv_sec);
   long usecdiff = (long)(tv1->tv_usec - tv2->tv_usec);
   return secdiff? secdiff: usecdiff;
}

//----------------------------------------------------------------------------

/** Add a check result to the dnx check result list in sorted order
 *
 * The check result is added to the check result list in ascending order, 
 * sorted by finish time.
 *
 * @param[in] cr - the check result to be addd to the results list.
 */
static void dnxAddResultToList(check_result * newcr)
{
   check_result ** curp;

   assert(newcr);

   DNX_PT_MUTEX_LOCK(&dnxResultListMutex);

   for (curp = &dnxResultList; *curp; curp = &(*curp)->next)
      if (dnxTimeCompare(&(*curp)->finish_time, &newcr->finish_time) >= 0)
         break;

   newcr->next = *curp;
   *curp = newcr;

   DNX_PT_MUTEX_UNLOCK(&dnxResultListMutex);
}

//----------------------------------------------------------------------------

/** Post job result information to Nagios 3.x.
 *
 * This version of post-results bypasses the Nagios 3.x filesystem-based 
 * results posting mechanism in favor of posting results directly to the
 * linked results list that the results reaper function fills with results 
 * from the results queue.
 *
 * @param[in] svc - the nagios service to which the results belong.
 * @param[in] check_type - nagios 3.x check type value.
 * @param[in] check_options - nagios 3.x bit-wise check options.
 * @param[in] schedule - boolean; nagios 3.x schedule flag.
 * @param[in] reschedule - boolean; nagios 3.x reschedule flag.
 * @param[in] latency - nagios 3.x latency value.
 * @param[in] start_time - the check start time in seconds.
 * @param[in] finish_time - the check finish time in seconds.
 * @param[in] early_timeout - boolean; true (!0) means the service timed out.
 * @param[in] exited_ok - boolean; true (!0) if the external check exited ok.
 * @param[in] res_code - the check result code.
 * @param[in] res_data - the check result data.
 *
 * @return Zero on success, or a non-zero error value.
 */
static int nagios3xPostResult(service * svc, int check_type,
      int check_options, int schedule, int reschedule, 
      time_t start_time, time_t finish_time, int early_timeout,
      int exited_ok, int res_code, char * res_data)
{
   check_result * newcr;

   // use malloc here, not xmalloc - nagios will free it
   if ((newcr = (check_result *)malloc(sizeof *newcr)) == 0)
      return DNX_ERR_MEMORY;

   // initialize and fill with result info
   newcr->object_check_type = SERVICE_CHECK;
   newcr->host_name = strdup(svc->host_name);
   newcr->service_description = strdup(svc->description);
   newcr->check_type = check_type;
   newcr->check_options = check_options;
   newcr->scheduled_check = schedule;
   newcr->reschedule_check = reschedule;
   newcr->output_file = 0;
   newcr->output_file_fd = -1;
   newcr->latency = svc->latency;
   newcr->start_time.tv_sec = start_time;
   newcr->start_time.tv_usec = 0;
   newcr->finish_time.tv_sec = finish_time;
   newcr->finish_time.tv_usec = 0;
   newcr->early_timeout = early_timeout;
   newcr->exited_ok = exited_ok;
   newcr->return_code = res_code;
   newcr->output = strdup(res_data);
   newcr->next = 0;

   dnxAddResultToList(newcr);

   return 0;
}

//----------------------------------------------------------------------------

/** Merge two sorted lists into a single list.
 *
 * @param[in] lista - the first list to be merged.
 * @param[in] listb - the second list to be merged.
 * @return a combination of @p lista and @p listb. 
 */
static check_result * dnxMergeLists(check_result * lista, check_result * listb)
{
   check_result * result = 0;

   check_result ** iter;
   for (iter = &result; lista && listb; iter = &(*iter)->next)
      if (dnxTimeCompare(&lista->finish_time, &listb->finish_time) <= 0)
      { *iter = lista; lista = lista->next; }
      else
      {*iter = listb; listb = listb->next; }

   *iter = lista? lista: listb;

   return result;
}

//----------------------------------------------------------------------------

/** Move the DNX N3 results queue to Nagios.
 *
 * This function should only be called on the Nagios timed event handler
 * thread so there are no race conditions between Nagios's processing of it's 
 * results queue and DNX's addition of data to that queue. 
 */
static void dnxMoveResultsToNagios(void)
{
   check_result * local;

   // safely save off currently local list
   DNX_PT_MUTEX_LOCK(&dnxResultListMutex);
   local = dnxResultList;
   dnxResultList = 0;
   DNX_PT_MUTEX_UNLOCK(&dnxResultListMutex);

   // merge local into check_result_list, store in check_result_list
   check_result_list = dnxMergeLists(local, check_result_list);
}

//----------------------------------------------------------------------------

/** Timed Event Handler.
 *
 * Nagios calls this routine once each time a timed event needs to execute. 
 * The particular event we care about is the REAPER event. 
 *
 * @param[in] event_type - the event type for which we're being called.
 * @param[in] data - an opaque pointer to nagios event-specific data.
 *
 * @return Zero, but the value is ignored by Nagios in this event.
 */
static int ehTimedEvent(int event_type, void * data)
{
   nebstruct_timed_event_data * ted = (nebstruct_timed_event_data *)data;
   timed_event * event = (timed_event*)data;
   int ret;

   // sanity checks
   if (event_type != NEBCALLBACK_TIMED_EVENT_DATA || ted == 0)
      return ERROR;

   // we only care about REAPER events
   if (ted->event_type != EVENT_CHECK_REAPER)
      return OK;

   dnxDebug(3, "Reaper handler called.");
   
   dnxMoveResultsToNagios();

   return OK;
}

#else // !DIRECT_POST

//----------------------------------------------------------------------------

/** Move a check result temporary file to the queue directory.
 *
 * This is a modified copy of a Nagios 3.x function. The difference is this 
 * version doesn't set umask due to race conditions with nagios core code.
 *
 * @param[in] resfile - the result file to be moved.
 */
static void nagios3xMoveCheckResultToQueue(char * resfile)
{
   char * outfile;
   int ret = -1;

   // a nagios 3x global variable
   extern char * check_result_path;

   // create a safe temp file
   // path length + 'cXXXXXX' (7) + '/' (1) + '.ok' (3) + null (1)
   if ((outfile = (char *)xmalloc(strlen(check_result_path) 
         + 7 + 1 + 3 + 1)) != 0)
   {
      int fd;
      //int oldmask = umask(077);

      sprintf(outfile, "%s/cXXXXXX", check_result_path);
      if ((fd = mkstemp(outfile)) > 0)
      {
#ifdef __CYGWIN__
         // Cygwin cannot rename open files - gives Permission Denied
         close(fd);
#endif

         // move the original file using a nagios core function
         ret = my_rename(resfile, outfile);

#ifndef __CYGWIN__
         // close the file
         close(fd);
#endif

         // create an ok-to-go indicator file
         strcat(outfile, ".ok");
         if ((fd = open(outfile, O_CREAT|O_WRONLY|O_TRUNC, S_IRUSR|S_IWUSR)) > 0)
            close(fd);

      }

      //umask(oldmask);

      xfree(outfile);
   }

   // delete the original file if allocation failed or it couldn't be moved
   if (ret != 0)
      unlink(resfile);
}

//----------------------------------------------------------------------------

/** Post job result information to Nagios 3.x.
 *
 * Nagios 3.x uses a filesystem-based method of posting results. Results are
 * written to a temporary file. Once the temp file is closed, it's moved into
 * the Nagios results queue directory, and a marker file of the same name, but
 * with a ".ok" suffice, is created. This marker file tells Nagios that the 
 * results file is complete and ready to be processed.
 *
 * Nagios writes the first half of the results file in a temporary directory, 
 * and allows the child check process to finish it and move it to the results
 * queue directory. DNX foregoes the temporarly file and writes the results 
 * file directly into the queue directory, and then creates the marker file.
 *
 * SECURITY NOTE: Nagios creates the results file in a secure manner by 
 * setting the umask to mask off ALL rights from anyone but the owner of the 
 * file (Nagios). DNX doesn't have the luxury of being able to manipulate the
 * umask because DNX check results are posted in a secondary thread, which is
 * not allowed to manipulate the umask because it would cause a race condition
 * with the main thread in Nagios.
 *
 * @param[in] svc - the nagios service to which the results belong.
 * @param[in] check_type - nagios 3.x check type value.
 * @param[in] check_options - nagios 3.x bit-wise check options.
 * @param[in] schedule - boolean; nagios 3.x schedule flag.
 * @param[in] reschedule - boolean; nagios 3.x reschedule flag.
 * @param[in] latency - nagios 3.x latency value.
 * @param[in] start_time - the check start time in seconds.
 * @param[in] finish_time - the check finish time in seconds.
 * @param[in] early_timeout - boolean; true (!0) means the service timed out.
 * @param[in] exited_ok - boolean; true (!0) if the external check exited ok.
 * @param[in] res_code - the check result code.
 * @param[in] res_data - the check result data.
 *
 * @return Zero on success, or a non-zero error value.
 *
 * @todo This routine should be in nagios code.
 */
static int nagios3xPostResult(service * svc, int check_type,
      int check_options, int schedule, int reschedule, 
      time_t start_time, time_t finish_time, int early_timeout,
      int exited_ok, int res_code, char * res_data)
{
   char * filename;
   FILE * fp = 0;
   int fd;

   // open a file in the check results path for storing check results
   // path length + 'checkXXXXXX' (11) + '/' (1) + '.ok' (3) + null (1)
   if ((filename = (char *)xmalloc(strlen(temp_path) 
         + 11 + 1 + 3 + 1)) == 0)
      return DNX_ERR_MEMORY;

   sprintf(filename, "%s/checkXXXXXX", temp_path);
   if ((fd = mkstemp(filename)) < 0
         || (fp = fdopen(fd, "w")) == 0)
   {
      xfree(filename);
      if (fd >= 0) close(fd);
      return DNX_ERR_OPEN;
   }

   // write check results to the queue file
   fprintf(fp, "### Active Check Result File ###\n");
   fprintf(fp, "file_time=%lu\n\n", (unsigned long)start_time);
   fprintf(fp, "### Nagios Service Check Result ###\n");
   fprintf(fp, "# Time: %s", ctime(&start_time));
   fprintf(fp, "host_name=%s\n", svc->host_name);
   fprintf(fp, "service_description=%s\n", svc->description);
   fprintf(fp, "check_type=%d\n", check_type);
   fprintf(fp, "check_options=%d\n", check_options);
   fprintf(fp, "scheduled_check=%d\n", schedule);
   fprintf(fp, "reschedule_check=%d\n", reschedule);
   fprintf(fp, "latency=%f\n", svc->latency);
   fprintf(fp, "start_time=%lu.0\n", (unsigned long)start_time);
   fprintf(fp, "finish_time=%lu.%lu\n", (unsigned long)finish_time);
   fprintf(fp, "early_timeout=%d\n", early_timeout);
   fprintf(fp, "exited_ok=%d\n", exited_ok);
   fprintf(fp, "return_code=%d\n", res_code);
   fprintf(fp, "output=%s\n", res_data);

   fclose(fp);

   nagios3xMoveCheckResultToQueue(filename);

   xfree(filename);

   return 0;
}

#endif // ?DIRECT_POST

#else
# error Unsupported NEB API version.
#endif

//----------------------------------------------------------------------------

static int dnxPostResult(void * data, time_t start_time, unsigned delta,
      int early_timeout, int res_code, char * res_data)
{
   DnxJobTransfer * job = (DnxJobTransfer *)data;
   service * svc = (service *)job->svc;
   int ret;

   assert(job);

   /** @todo Investigate the rationale for setting res_code this way. */

   if (early_timeout)
      res_code = STATE_UNKNOWN;

#if CURRENT_NEB_API_VERSION == 2

   ret = nagios2xPostResult(svc, start_time, early_timeout,
         res_code, res_data);

#elif CURRENT_NEB_API_VERSION == 3

   /** @todo Nagios 3.x: Collect a better value for exited_ok. */
   /** @todo Nagios 3.x: Collect a better value for check_type. */

   ret = nagios3xPostResult(svc, SERVICE_CHECK_ACTIVE,
         job->chkopts, job->schedule, job->reschedule,
         start_time, start_time + delta, early_timeout,
         1, res_code, res_data);

#else
# error Unsupported NEB API version.
#endif

   xfree(job); // free the originally allocated DnxJobTransfer structure

   return ret;
}

//----------------------------------------------------------------------------

/** Retrieve and validate the next message's header.
 *
 * @param[in] phdr - the header buffer to be filled.
 *
 * @return zero on success or a non-zero error code.
 */
static int GetMsgHeader(DnxMsgHeader * phdr)
{
   size_t toread = sizeof *phdr;
   char * bufptr = (char *)phdr;
   ssize_t bytesread;

   // read the message header from the plugin module
   while (toread > 0 && (bytesread = read(indatafd, bufptr, toread)) > 0)
   {
      toread -= bytesread;
      bufptr += bytesread;
   }

   // check for closed pipe
   if (bytesread == 0)
      return EPIPE;

   // check for error
   if (bytesread < 0)
   {
      int ret = errno;

      /** @todo EPIPE means broken pipe: restart child process */

      dnxDebug(1, "GetMsgHeader: read failed: %s.", strerror(ret));
      return ret;
   }

   // sanity check message format
   if (phdr->signature != DNX_MSG_SIGNATURE
         || phdr->structsz > DNX_MAX_XFER_SIZE)
   {
      dnxDebug(1, "GetMsgHeader: unsupported message format.");
      return DNX_ERR_UNSUPPORTED;
   }

   return 0;
}

//----------------------------------------------------------------------------

/** Transfer a result from the server and post it to Nagios.
 *
 * @param[in,out] result - the current result structure pointer. 
 *
 * @return zero on success or a non-zero error code.
 */
static int ProcessResultTransferMsg(DnxResultTransfer * result)
{
   size_t toread = result->hdr.structsz - sizeof result->hdr;
   char * bufptr = (char *)result + sizeof result->hdr;
   ssize_t bytesread;
   int ret;

   // read the rest of the result transfer structure
   while (toread > 0 && (bytesread = read(indatafd, bufptr, toread)) > 0)
   {
      bufptr += bytesread;
      toread -= bytesread;
   }

   // check for closed pipe
   if (bytesread == 0)
      return EPIPE;

   // check for error 
   if (bytesread < 0)
   {
      ret = errno;

      /** @todo EPIPE means broken pipe: restart child process */

      dnxDebug(1, "ProcessResultTransferMsg: read failed: %s.", strerror(ret));
      return ret;
   }

   // post result to Nagios
   if ((ret = dnxPostResult(result->payload, result->start_time, result->delta,
         result->early_timeout, result->res_code, result->res_data)) != 0)
      dnxLog("Failed to post result [%lu] to Nagios; \"%s\": %d.", 
            result->serial, result->res_data, ret);

   return ret;
}

//----------------------------------------------------------------------------

/** Results listener thread.
 *
 * Listens for the server to post results. Captures each result and posts it
 * to the Nagios results queue.
 *
 * @param arg the thread routine argument.
 * @return the thread return value.
 */
static void * dnxResultsListener(void * arg) 
{
   DnxResultTransfer * result = 0;
   size_t resultsz = 0;
   int terminating = 0;

   // ignore broken pipe signals to this thread
   signal(SIGPIPE, SIG_IGN);
   
   dnxLog("Results listener thread awaiting results...");

   while (!terminating)
   {
      int ret;
      DnxMsgHeader hdr;
      if ((ret = GetMsgHeader(&hdr)) == 0)
      {
         switch (hdr.msgtype)
         {
            case DNX_MSG_RESULT_TRANSFER:
               // resize result transfer buffer if necessary
               if (hdr.structsz > resultsz)
               {
                  DnxResultTransfer * tmp;
                  if ((tmp = (DnxResultTransfer *)xrealloc(result, hdr.structsz)) != 0)
                  {
                     result = tmp;
                     resultsz = hdr.structsz;
                  }
               }

               // copy header and read rest of job transfer message
               result->hdr = hdr;
               ret = ProcessResultTransferMsg(result);
               break;
         }
      }
      if (ret == EPIPE)
         terminating = 1;
   }

   dnxLog("Results listener thread terminating...");

   xfree(result);
   return 0;
}

//----------------------------------------------------------------------------

/** Read a single ACK/NAK from the server.
 *
 * The ack/nak is actually a DNX error code (zero on success).
 *
 * @return an error, or the value of the ack/nak.
 */
static int GetAckNak(void)
{
   int ret;
   size_t toread = sizeof ret;
   char * bufptr = (char *)&ret;
   ssize_t bytesread;

   while (toread > 0 && (bytesread = read(inackfd, bufptr, toread)) > 0)
   {
      bufptr += bytesread;
      toread -= bytesread;
   }

   // check for closed pipe
   if (bytesread == 0)
      return EPIPE;

   // check for error
   if (bytesread < 0)
   {
      ret = errno;

      /** @todo EPIPE means broken pipe; restart server. */

      dnxDebug(1, "GetAckNak: read failed: %s.", strerror(ret));
      return ret;
   }

   return ret;
}

//----------------------------------------------------------------------------

/** Request a client node request from the server.
 *
 * Once a node request allocation has been successfully requested, it must be
 * followed (asap) by a job transfer because the server has one on hold and is
 * now waiting for the job to be transferred. Additionally, a node request has
 * a timeout, and too much delay will cause it to be replaced.
 *
 * @return Zero on success, or a non-zero error value.
 */
static int AllocateNodeRequest(void)
{
   DnxMsgHeader nodereq;
   size_t towrite = nodereq.structsz = sizeof nodereq;
   char * bufptr = (char *)&nodereq;
   ssize_t byteswritten;
   int ret;

   dnxDebug(3, "Allocating node request.");

   // allocate client node request - bail on EAGAIN - pipe full, no one's listening
   nodereq.signature = DNX_MSG_SIGNATURE;
   nodereq.msgtype = DNX_MSG_RESERVE_NODEREQ;

   while (towrite > 0 && (byteswritten = write(outdatafd, bufptr, towrite)) > 0)
   {
      bufptr += byteswritten;
      towrite -= byteswritten;
   }

   // check for closed pipe
   if (byteswritten == 0)
      return EPIPE;

   // check for error
   if (byteswritten < 0)
   {
      ret = errno;

      /** @todo EPIPE means broken pipe; restart server. */

      dnxDebug(1, "Alloc node request: write failed: %s.", strerror(ret));
      return ret;
   }

   return GetAckNak();
}

//----------------------------------------------------------------------------

/** Transfer a job to the dnx server.
 *
 * A job transfer request must be preceded by an allocate client node request.
 *
 * The request is larger than an integer, so we assume it can take more than 
 * one pass, but the result is an int, so we hope we can read it in one pass.
 *
 * @param[in] job - the job to be transferred to the server.
 *
 * @return zero on success or a non-zero error code.
 */
static int TransferJobToServer(DnxJobTransfer * job)
{
   char * bufptr = (char *)job;
   size_t towrite = job->hdr.structsz;
   ssize_t byteswritten;
   int ret;

   assert(job);

   dnxDebug(3, "Transferring service check.");

   // write job to server - bail on EAGAIN - pipe full, no one's listening
   job->hdr.signature = DNX_MSG_SIGNATURE;
   job->hdr.msgtype = DNX_MSG_JOB_TRANSFER;
   while (towrite > 0 && (byteswritten = write(outdatafd, bufptr, towrite)) > 0)
   {
      bufptr += byteswritten;
      towrite -= byteswritten;
   }

   // check for closed socket
   if (byteswritten == 0)
      return EPIPE;

   // check for error
   if (byteswritten < 0)
   {
      ret = errno;

      /** @todo EPIPE means broken pipe; restart server. */

      dnxDebug(1, "Alloc node request: write failed: %s.", strerror(ret));
      return ret;
   }

   return GetAckNak();
}

//----------------------------------------------------------------------------

/** Post a new job from Nagios to the dnx server.
 *
 * The payload of the job transfer structure is the address of the structure
 * itself. This structure is allocated here, but freed when the results are
 * collected and posted. If no result is posted, the structure address is 
 * lost.
 *
 * @param[in] serial - the serial number of the new job.
 * @param[in] ds - a pointer to the nagios job that's being posted.
 *
 * @return Zero on success, or a non-zero error value.
 *
 * @todo Track posted jobs and free them if associated result is lost.
 */
static int dnxPostNewJob(unsigned long serial, nebstruct_service_check_data * ds)
{
   service * svc = (service *)ds->OBJECT_FIELD_NAME;
   DnxJobTransfer * job;
   size_t jobsz;
   int ret;

   assert(ds);
   assert(svc);
   assert(ds->command_line);

   // tell server to allocate a node request
   if ((ret = AllocateNodeRequest()) != 0)
      return ret;

   // allocate a job transfer structure
   jobsz = sizeof(*job) + strlen(ds->command_line);
   if ((job = (DnxJobTransfer *)xmalloc(jobsz)) == 0)
      return DNX_ERR_MEMORY;

   job->hdr.structsz = jobsz;
   job->serial = serial;
   job->payload = job;  // track job transfer structure between job and result
   job->svc = svc;      // track service for use in result posting 
   job->start_time = ds->start_time.tv_sec;
   job->timeout = ds->timeout;
   job->expires = job->start_time + job->timeout + 5;

#if CURRENT_NEB_API_VERSION == 3
   {
      // a nagios 3.x global variable
      extern check_result check_result_info;

      /** @todo patch nagios to pass these values to the event handler. */

      job->chkopts = check_result_info.check_options;
      job->schedule = check_result_info.scheduled_check;
      job->reschedule = check_result_info.reschedule_check;
   }
#endif

   strcpy(job->cmd, ds->command_line);

   return TransferJobToServer(job);
} 

//----------------------------------------------------------------------------

/** Service Check Event Handler.
 *
 * Nagios calls this routine multiple times for each service check. We only 
 * care about the INITIATE call, so we return OK for all others.
 *
 * @param[in] event_type - the event type for which we're being called.
 * @param[in] data - an opaque pointer to nagios event-specific data.
 *
 * @return Zero if we want Nagios to handle the event;
 *    NEBERROR_CALLBACKOVERRIDE indicates that we want to handle the event
 *    ourselves; any other non-zero value represents an error.
 */
static int ehSvcCheck(int event_type, void * data)
{
   static unsigned long serial = 0;

   nebstruct_service_check_data * svcdata = (nebstruct_service_check_data *)data;
   int ret;

   // shouldn't happen because we didn't sign up for others
   if (event_type != NEBCALLBACK_SERVICE_CHECK_DATA)
      return OK;

   // if this happens it's a bug in Nagios
   if (svcdata == 0)
   {
      dnxDebug(1, "Service handler received NULL service data structure.");
      return ERROR;  // shouldn't happen - internal Nagios error
   }

   // we only want the first BROKER event we get on each check
   if (svcdata->type != NEBTYPE_SERVICECHECK_INITIATE)
      return OK;     // ignore non-initiate service checks

   // every valid service check is assigned a unique serial number
   serial++;

   dnxDebug(3, "Service handler received service check [%lu] at %lu.",
         serial, (unsigned long)svcdata->start_time.tv_sec);

   // check for local execution pattern on command line
   if (cfg.localCheckPattern 
         && regexec(&regEx, svcdata->command_line, 0, 0, 0) == 0)
   {
      dnxDebug(1, "Regex filter. Service check [%lu] will execute locally: %s.", 
            serial, svcdata->command_line);
      return OK;
   }

   // try to post this job to the dnx server process
   if ((ret = dnxPostNewJob(serial, svcdata)) != DNX_OK)
   {
      dnxDebug(1, "Post failed: %s. Service check [%lu] will execute locally: %s.", 
            dnxErrorString(ret), serial, svcdata->command_line);
      return OK;
   }

   return NEBERROR_CALLBACKOVERRIDE;   // tell nagios we want it
}

//----------------------------------------------------------------------------

/** Calculate the optimal size of the job list.
 *
 * Assumes the caller will actually use the returned value to allocate the
 * job list. Based on this assumption, this routine logs messages indicating
 * when various configuration overrides have taken effect.
 *
 * @return The calculated size of the job list.
 */
static int dnxCalculateJobListSize(void)
{
   int size = nagiosGetServiceCount();

   // zero doesn't make sense...
   if (size < 1)
   {
      size = 100;
      dnxLog("No Nagios services defined! "
             "Defaulting to %d slots in the DNX job queue.", size);
   }
   return size;
}

//----------------------------------------------------------------------------

/** Exec a server child process with comm channels on well-known descriptors.
 *
 * The child process is an entirely different program, so we have to dup the
 * pipe descriptors onto well-known descriptor numbers so the server will 
 * know where to find them. I considered using STDIO channels for this, but
 * the server may want to use STDIO for other reasons. Too bad, because it's
 * a perfect fit: STDIN = job xfer, STDOUT = ack/nak, and STDERR = results.
 * If I can find a way to keep the user from configuring any logging on STDIO,
 * I may yet convert over. The handle values are defined in the plugin main 
 * header file, so it would be simple to redefine those descriptor values.
 *
 * @return Zero on success or a non-zero error code.
 */
static int execServerProcess(void)
{

#define P0R 0     // Pipe 0 Read descriptor
#define P0W 1     // Pipe 0 Write descriptor
#define P1R 2     // Pipe 1 Read descriptor
#define P1W 3     // Pipe 1 Write descriptor
#define P2R 4     // Pipe 2 Read descriptor
#define P2W 5     // Pipe 2 Write descriptor
#define PMX 6     // Pipe Max desscriptors

   pid_t pid;
   int flags, fds[PMX];

   // open three IPC pipes and fork the process
   memset(fds, -1, sizeof fds);
   if (pipe(fds + P0R) < 0 || pipe(fds + P1R) < 0 
         || pipe(fds + P2R) < 0 || (pid = fork()) < 0)
   {
      int i, ret = errno;

      for (i = 0; i < elemcount(fds); i++)
         if (fds[i] >= 0) 
            close(fds[i]);

      return ret;
   }

   // *** zero is child process ***

   if (pid == 0)
   {
      int svclistsz = dnxCalculateJobListSize();
      char strsvclistsz[16];

      // create a string version of the service list size
      snprintf(strsvclistsz, sizeof strsvclistsz, "%d", svclistsz);

      close(fds[P0W]);     // job xfer    - server closes write end
      close(fds[P1R]);     // ack/nak     - server closes read end
      close(fds[P2R]);     // result xfer - server closes read end
      
      // dup open pipe descriptors onto well-known descriptors, if needed
      if (fds[P0R] != DNX_SRVFD)
      {
         dup2(fds[P0R], DNX_SRVFD);
         close(fds[P0R]); 
      }
      if (fds[P1W] != DNX_SAKFD)
      {
         dup2(fds[P1W], DNX_SAKFD);
         close(fds[P1W]);
      }
      if (fds[P2W] != DNX_PLGFD)
      {
         dup2(fds[P2W], DNX_PLGFD);
         close(fds[P2W]);
      }

      // exec DNX Server process, pass config args
      execl(cfg.serverPath, cfg.serverPath, "-c", cfgfile, 
            "-j", strsvclistsz, (char *)0);

      _exit(1);            // error in exec - kill child

      // *** NO RETURN FROM HERE ***
   }

   // *** otherwise we're still in the parent process ***

   close(fds[P0R]);        // job xfer    - plugin closes read end
   close(fds[P1W]);        // ack/nak     - plugin closes write end
   close(fds[P2W]);        // result xfer - plugin closes write end

   // save open pipe descriptors in global variables
   outdatafd = fds[P0W];
   inackfd   = fds[P1R];
   indatafd  = fds[P2R];

   // set non-block flag on plugin -> server write channel
   // if the pipe is full, we'll let Nagios handle the check
   if ((flags = fcntl(outdatafd, F_GETFL, 0)) >= 0)
   {
      flags |= O_NONBLOCK;
      fcntl(outdatafd, F_SETFL, flags);
   }

   // save server pid for wait call
   serverpid = pid;

   return 0;
}

//----------------------------------------------------------------------------

/** Terminate an existing server process.
 *
 * @return Zero on success or a non-zero error code.
 */
static int terminateServerProcess(void)
{
   int ret = 0;
   if (serverpid > 0)
   {
      kill(serverpid, SIGTERM);  // sigterm stops agent thread
      close(outdatafd);          // closing pipes stop request handler
      close(inackfd);
      if (waitpid(serverpid, 0, 0) < 0)
         ret = errno;
   }
   return ret;
}

//----------------------------------------------------------------------------

// forward declaration due to circular reference
static int ehProcessData(int event_type, void * data);

/** Deinitialize the DNX Plugin.
 *
 * @return Always returns zero.
 */
static int dnxPluginDeInit(void)
{
   // deregister for all nagios events we previously registered for...
   neb_deregister_callback(NEBCALLBACK_PROCESS_DATA, ehProcessData);
   neb_deregister_callback(NEBCALLBACK_SERVICE_CHECK_DATA, ehSvcCheck);
#if CURRENT_NEB_API_VERSION == 3 && defined(DIRECT_POST)
   neb_deregister_callback(NEBCALLBACK_TIMED_EVENT_DATA, ehTimedEvent);
#endif

   // shutdown server child process
   if (serverpid != 0)
      terminateServerProcess();

   // terminate results listener thread
   if (rlthread != 0)
   {
      close(indatafd);           // closing pipe signals thread
      pthread_join(rlthread, 0);
   } 

   dnxLog("DNX Nagios Plugin shutdown completed.");

   return OK;  // Nagios OK value
}

//----------------------------------------------------------------------------

/** Initialize the DNX plugin.
 *
 * @return Zero on success, or a non-zero error value.
 */
static int dnxPluginInit(void)
{
   int ret;

   // start dnx server child process
   if ((ret = execServerProcess()) != 0)
   {
      dnxLog("Error starting DNX server process: %s.", dnxErrorString(ret));
      return ret;
   }

   dnxLog("DNX Server process started.");

   // start results listener
   if ((ret = pthread_create(&rlthread, 0, dnxResultsListener, (void *)0)) != 0)
      return ret;
   
   dnxLog("Results listener thread started.");

#if CURRENT_NEB_API_VERSION == 3 && defined(DIRECT_POST)

   // register for timed event to piggy-back on reaper thread
   neb_register_callback(NEBCALLBACK_TIMED_EVENT_DATA, myHandle, 0, ehTimedEvent);
   dnxLog("Registered for TIMEDEVENT_EXECUTE event.");

#endif

   // registration for this event starts everything rolling
   neb_register_callback(NEBCALLBACK_SERVICE_CHECK_DATA, myHandle, 0, ehSvcCheck);
   dnxLog("Registered for SERVICE_CHECK_DATA event.");

   dnxLog("DNX Nagios Plugin initialization completed.");

   return OK;  // Nagios OK value
}

//----------------------------------------------------------------------------

/** Launches an external command and waits for it to return a status code.
 *
 * @param[in] script - the command line to be launched.
 *
 * @return Zero on success, or a non-zero error value.
 */
static int launchScript(char * script)
{
   int ret;

   assert(script);

   // exec the script - system waits till child completes
   if ((ret = system(script)) == -1)
   {
      dnxLog("Failed to exec script: %s.", strerror(errno));
      ret = DNX_ERR_INVALID;
   }
   else
      ret = DNX_OK;

   dnxLog("Sync script returned %d.", WEXITSTATUS(ret));

   return ret;
}

//----------------------------------------------------------------------------

/** Process Data Event Handler.
 *
 * We actually register for service check events and start the child process
 * server program in this event handler because Nagios starts NEB modules 
 * long before it's fully configured its service check list. By waiting to 
 * initialize the server in this event handler, we have access to the number
 * of service checks that Nagios is configured to manage, and we can thus
 * properly configure the DNX server child process with a properly sized 
 * job queue.
 *
 * @param[in] event_type - the event regarding which we were called by Nagios.
 * @param[in] data - an opaque pointer to an event-specific data structure.
 *
 * @return Zero if all is okay, but we want nagios to handle this event;
 *    non-zero if there's a problem of some sort.
 */
static int ehProcessData(int event_type, void * data)
{
   nebstruct_process_data *procdata = (nebstruct_process_data *)data;

   // validate our event type - ignore wrong event type
   assert(event_type == NEBCALLBACK_PROCESS_DATA);
   if (event_type != NEBCALLBACK_PROCESS_DATA)
      return OK;

   // sanity-check our data structure - should never happen
   if (!procdata)
   {
      dnxLog("Startup handler received NULL process data structure.");
      return ERROR;
   }

   // look for process event loop start event
   if (procdata->type == NEBTYPE_PROCESS_EVENTLOOPSTART)
   {
      dnxDebug(2, "Startup handler received PROCESS_EVENTLOOPSTART event.");

      // execute sync script, if defined
      if (cfg.syncScript)
      {
         dnxLog("Startup handler executing plugin sync script: %s.", cfg.syncScript);

         // NB: This halts Nagios execution until the script exits...
         launchScript(cfg.syncScript);
      }

      // if plugin init fails, do plugin shutdown
      if (dnxPluginInit() != 0)
         dnxPluginDeInit();
   }
   return OK;
}

/*--------------------------------------------------------------------------
                                 INTERFACE
  --------------------------------------------------------------------------*/

//----------------------------------------------------------------------------

/** The main NEB module deinitialization routine.
 *
 * This function gets called when the module is unloaded by the event broker.
 *
 * @param[in] flags - nagios NEB module flags - not used.
 * @param[in] reason - nagios reason code - not used.
 *
 * @return Always returns zero.
 */
int nebmodule_deinit(int flags, int reason)
{
   dnxLog("-------- DNX Nagios Plugin Module Shutdown Initiated --------");
   
   dnxPluginDeInit();
   releaseConfig();
   xheapchk();

   dnxLog("-------- DNX Nagios Plugin Module Shutdown Completed --------");
   return 0;
}

//----------------------------------------------------------------------------

/** The main NEB module initialization routine.
 *
 * This function gets called when the module is loaded by the event broker.
 *
 * @param[in] flags - module flags - not used
 * @param[in] args - module arguments. These come from the nagios
 *    configuration file, and are passed through to the module as it loads.
 * @param[in] handle - our module handle - passed from the OS to nagios as
 *    nagios loaded us.
 *
 * @return Zero on success, or a non-zero error value.
 */
int nebmodule_init(int flags, char * args, nebmodule * handle)
{
   int ret;

   myHandle = handle;

   // module args string should contain a fully-qualified config file path
   if (!args || !*args)
      args = DNX_DEFAULT_SERVER_CONFIG_FILE;

   cfgfile = args;
   if ((ret = initConfig(cfgfile)) != 0)
      return ERROR;

   // set configured debug level and syslog log facility code
   dnxLogInit(cfg.logFilePath, cfg.debugFilePath, cfg.auditFilePath, 
         &cfg.debugLevel);

   dnxLog("-------- DNX Nagios Plugin Module Version %s Startup --------", VERSION);
   dnxLog("Copyright (c) 2006-2010 Intellectual Reserve. All rights reserved.");
   if (cfg.debugLevel)
      dnxLog("Debug logging enabled at level %d to %s.",
            cfg.debugLevel, cfg.debugFilePath);
   if (cfg.auditFilePath)
      dnxLog("Auditing enabled to %s.", cfg.auditFilePath);

   // subscribe to PROCESS_DATA call-backs in order to defer initialization
   //    until after Nagios validates its configuration and environment.
   if ((ret = neb_register_callback(NEBCALLBACK_PROCESS_DATA,
         myHandle, 0, ehProcessData)) != OK)
   {
      dnxLog("PROCESS_DATA event registration failed: %s.", dnxErrorString(ret));
      terminateServerProcess();
      releaseConfig();
      return ERROR;
   }
   start_time = time(0);

   dnxLog("-------- DNX Nagios Plugin Module Startup Complete --------");

   return OK;
}

