/*--------------------------------------------------------------------------

   Copyright (c) 2006-2010, Intellectual Reserve, Inc. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 2 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  --------------------------------------------------------------------------*/

/** Implements statistics management functionality.
 *
 * This implementation uses no locking in order to reduce contention. Since 
 * no nodes are ever deleted, this shouldn't be a problem, as all access is
 * read-only. When adding nodes, there is a possibility that the same node
 * will be added multiple times by multiple threads. To avoid this issue, 
 * ensure that the first stat increment of each new node is always done by 
 * a single thread.
 *
 * This implementation is based on a hash table of fixed size containing
 * linked bucket lists. Because the lists are singly-linked, insertion is
 * atomic. A reader will either see a node or not, depending on if the read
 * happened to take place after or before the atomic write of the previous
 * nodes "next" field.
 *
 * @file dnx/plugin/server/dnxStats.c
 * @author Steven D. Morrey (dnx-devel@lists.sourceforge.net)
 * @author John Calcote (dnx-devel@lists.sourceforge.net)
 * @attention Please submit patches to http://dnx.sourceforge.net
 * @ingroup DNX_SERVER_IMPL
 */

#include "dnxStats.h"
#include "dnxDebug.h"
#include "dnxError.h"

#include <pthread.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#define elemcount(x) (sizeof(x)/sizeof(*(x)))

/** The structure for managing stats for one client node, by IP address. */
typedef struct DnxNode
{
   struct DnxNode * next;     //!< Next node in bucket list.
   struct sockaddr * saddr;   //!< data.addr cast to struct sockaddr.
   DnxNodeData data;          //!< The data for this node.
} DnxNode;

/* Global (static) variables managed by this module. */
static DnxNode * s_nodeHashTable[256];       //!< The global node hash table.
static unsigned s_serverStats[STATS_COUNT];  //!< The global server stats table.

/*--------------------------------------------------------------------------
                              IMPLEMENTATION
  --------------------------------------------------------------------------*/

/** Return a 1-byte hash code for the specified address. 
 *
 * @param[in] addr - the address to be hashed.
 *
 * @return an unsigned hash code with a maximum value of 255.
 * 
 * @note: This hash function assumes that @p addr is at LEAST 4 bytes long.
 */
static unsigned hashFunction(struct sockaddr * addr)
{
   unsigned hash = 0;
   switch (addr->sa_family)
   {
      case AF_INET:
         hash = ((struct sockaddr_in *)addr)->sin_addr.s_addr;
         break;

      case AF_INET6:
      {
         int i;
         for (i = 0; i < 4; i++)
            hash += *((unsigned *)(((struct sockaddr_in6 *)addr)
                  ->sin6_addr.s6_addr + i * 4)) & 0xFFFFFFFF;
      }

      default:
         hash = addr->sa_data[2];   // try to skip the port
         break;
   }
   return hash & 0xFF;
}

//----------------------------------------------------------------------------

/** Compare two addresses.
 *
 * Only the network address portion of the address is compared;
 * port numbers are ignored.
 *
 * @param[in] addr_a - the first address to compare
 * @param[in] addr_b - the second address to compare
 * 
 * @return A value less than zero if @p addr_a is less than @p addr_b.
 *         Zero if @p addr_a is equal to @p addr_b.
 *         A value greater than zero if @p addr_a is greater than @p addr_b.
 */
static int dnxAddrCompare(struct sockaddr * addr_a, struct sockaddr * addr_b)
{
   if (addr_a->sa_family != addr_b->sa_family)
      return addr_a->sa_family - addr_b->sa_family;

   switch(addr_a->sa_family)
   {
      case AF_INET:
         return (int)(((struct sockaddr_in *)addr_a)->sin_addr.s_addr 
               - ((struct sockaddr_in *)addr_b)->sin_addr.s_addr);

      case AF_INET6:
         return memcmp(((struct sockaddr_in6 *)addr_a)->sin6_addr.s6_addr,
               ((struct sockaddr_in6 *)addr_b)->sin6_addr.s6_addr, 16);
   }

   return memcmp(addr_a->sa_data + 2, addr_b->sa_data + 2, 
         sizeof(addr_a->sa_data) - 2);
}

//----------------------------------------------------------------------------

/** Create, insert, and return a new node with a specified address and hash.
 *
 * The new node is inserted at the specified location in the hash table.
 *
 * @param[in] npp - the location at which to store the new node.
 * @param[in] saddr - the address for which a new node should be created.
 * @param[in] hashCode - the previously calculated hash code for @p addr.
 *
 * @return the new node, or zero on memory allocation failure.
 */
static DnxNode * dnxCreateNodeAt(DnxNode ** npp, struct sockaddr * saddr, 
      unsigned hashCode)
{
   DnxNode * np;

   // allocate and populate a new node
   if ((np = (DnxNode *)xmalloc(sizeof *np)) == 0)
      return 0;

   memset(np, 0, sizeof *np);
   memcpy(np->data.address, saddr, sizeof np->data.address);
   np->saddr = (struct sockaddr *)np->data.address;
   dnxNtop(saddr, np->data.addrstr, sizeof np->data.addrstr);

   return *npp = np;
}

//----------------------------------------------------------------------------

/** Return a node matching an address.
 * 
 * NO LOCKS: Ensure caller is a single thread. (At least for the create case.)
 *
 * Insert a new node and return it if missing.
 *
 * @param[in] saddr - the address of the node to be returned.
 *
 * @return a (possibly new) node matching the specified address.
 */
static DnxNode * dnxGetNode(struct sockaddr * saddr)
{
   DnxNode ** npp, * np;
   unsigned hashCode = hashFunction(saddr);

   for (npp = &s_nodeHashTable[hashCode]; 
         (np = *npp) && dnxAddrCompare(np->saddr, saddr); npp = &np->next)
      ;

   return np? np: dnxCreateNodeAt(npp, saddr, hashCode);
}

/*--------------------------------------------------------------------------
                                 INTERFACE
  --------------------------------------------------------------------------*/

void dnxStatsGetServerStats(unsigned * statsbuf)
{
   memcpy(statsbuf, s_serverStats, sizeof s_serverStats);
}

//----------------------------------------------------------------------------

void dnxStatsResetServerStats(void)
{
   memset(s_serverStats, 0, sizeof s_serverStats);
}

//----------------------------------------------------------------------------

void dnxStatsInc(char * addr, DnxStatsIndex member)
{
   if (addr)
   {
      DnxNode * np = dnxGetNode((struct sockaddr *)addr);
      assert(member >= 0 && member < STATS_COUNT);
      if (np) np->data.stats[member]++;
   }
   s_serverStats[member]++;
}

//----------------------------------------------------------------------------

int dnxStatsForEachNode(DnxStatsNodeCB * cb, void * data)
{
   int i, ret = 0;
   for (i = 0; i < elemcount(s_nodeHashTable); i++)
   {
      DnxNode * np;
      for (np = s_nodeHashTable[i]; np; np = np->next)
         if ((ret = cb(&np->data, data)) != 0)
            break;
   }
   return ret == DNX_CBCANCEL? 0: ret;
}

//----------------------------------------------------------------------------

int dnxStatsForNodeByAddrStr(char * addrstr, DnxStatsNodeCB * cb, void * data)
{
   struct addrinfo * aip, * curaip;
   DnxNode * np = 0;
   int ret;

   // convert addrstr to a sockaddr
   if ((ret = getaddrinfo(addrstr, 0, 0, &aip)) != 0)
      return ret;

   // search for the first node that matches one of the addresses found
   for (curaip = aip; !np && curaip; curaip = curaip->ai_next)
      if ((np = dnxGetNode(curaip->ai_addr)) != 0)
         ret = cb(&np->data, data);

   freeaddrinfo(aip);

   return ret;
}

//----------------------------------------------------------------------------

void dnxStatsCleanup()
{
   int i;
   for (i = 0; i < elemcount(s_nodeHashTable); i++)
   {
      DnxNode * np = s_nodeHashTable[i];

      while (np)
      {
         DnxNode * tmp = np->next;
         free(np);
         np = tmp;
      }
      s_nodeHashTable[i] = 0;
   }
}

//============================================================================

