/*--------------------------------------------------------------------------

   Copyright (c) 2006-2010, Intellectual Reserve, Inc. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 2 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  --------------------------------------------------------------------------*/

/** Provides the interface for the DNX stats management module.
 *
 * DNX node management includes the ability to efficiently track nodes by binary 
 * address, and to increment a set of statistics for each node. Access is read-
 * only except for incrementing and clearing stats. Nodes may be added but not 
 * removed. This is done so that locking may be minimized to reduce lock 
 * contention.
 *
 * @file dnx/plugin/server/dnxStats.h
 * @author Steven D. Morrey (dnx-devel@lists.sourceforge.net)
 * @author John Calcote (dnx-devel@lists.sourceforge.net) 
 * @attention Please submit patches to http://dnx.sourceforge.net
 * @ingroup DNX_SERVER_IMPL
 */

#ifndef _DNXSTATS_H_
#define _DNXSTATS_H_

#include "dnxTransport.h"

/** An enumeration of stats in the server stats structure. */
typedef enum DnxStatsIndex
{
   REQUESTS_RECEIVED = 0,           //!< Requests received (node/srv).
   REQUESTS_EXPIRED,                //!< Expired requests (node/srv).
   DISPATCHES_OK,                   //!< Successful dispatches (node/srv).
   DISPATCHES_FAILED,               //!< Failed dispatches (node/srv).
   RESULTS_OK,                      //!< Successful results received (node/srv).
   RESULTS_FAILED,                  //!< Failed results received (node/srv).
   RESULTS_TIMED_OUT,               //!< Timed out results received (node/srv).
   // new node/server stats should be added here

   JOBS_HANDLED,                    //!< Jobs handled by DNX (srv only).
   JOBS_REJECTED_NO_SLOTS,          //!< Jobs rejected because of no slots (srv only).
   JOBS_REJECTED_NO_NODES,          //!< Jobs rejected beause of no node requests (srv only).
   POST_RESULTS_OK,                 //!< Successful results posted to Nagios (srv only).
   POST_RESULTS_FAILED,             //!< Failed results posted to Nagios (srv only).
   // new server-only stats should be added here

   STATS_COUNT                      //!< Node stats available.
} DnxStatsIndex;
 
/** The data of a node. */
typedef struct DnxNodeData
{
   char address[DNX_MAX_ADDRESS];   //!< The node address.
   char addrstr[DNX_MAX_ADDRSTR];   //!< The node address as a string.
   unsigned stats[JOBS_HANDLED];    //!< The per-node stats array.
} DnxNodeData;

/** The node stats callback method signature.
 *
 * @param[in] node - the node data reference for this iteration.
 * @param[in] data - an opaque pointer pasSed from the interface method.
 *
 * @return 0 on success, non-zero to cancel callbacks. This value is
 *    returned to the interface method, unless it's DNX_CBCANCEL which
 *    is converted to success.
 */
typedef int DnxStatsNodeCB(DnxNodeData * node, void * data);

/** Return a current snapshot of the server stats array.
 *
 * The @p statsbuf array should be at least STATS_COUNT elements in length.
 * 
 * @param[out] statsbuf - the address of storage for all server stats.
 */
void dnxStatsGetServerStats(unsigned * statsbuf);

/** Reset all server stats to zero.
 */
void dnxStatsResetServerStats(void);

/** Increment a member value from a single node.
 *
 * @param[in] addr - the address of the node for which a stat should be
 *    incremented. This is the binary address, not the presentation address.
 * @param[in] member - the stat to be incremented.
 */
void dnxStatsInc(char * addr, DnxStatsIndex member);

/** For each node in the list, call a method and pass some data.
 *
 * @param[in] cb - the node processing function to be called.
 * @param[in] data - opaque data pointer passed to nodeProc.
 *
 * @return zero on success or callback-cancelled, or a non-zero 
 *    error code returned by @p nodeProc.
 */
int dnxStatsForEachNode(DnxStatsNodeCB * cb, void * data);

/** For the specified node, call a method and pass some data.
 *
 * @param[in] addrstr - the DNS or dotted decimal address of the node.
 * @param[in] cb - the node processing function to be called.
 * @param[in] data - opaque data pointer passed to nodeProc.
 *
 * @return zero on success or callback-cancelled, or a non-zero 
 *    error code returned by @p nodeProc.
 */
int dnxStatsForNodeByAddrStr(char * addrstr, DnxStatsNodeCB * cb, void * data);

/** A destructor for the stats management module.
 *
 * This function can be used to cleanup any resources allocated by this module
 * before shutdown.
 */
void dnxStatsCleanup();

#endif   /* _DNXSTATS_H_ */

